/*//////////////////////////////////////////////////////////////////////////
// BANNER PRINTING PROGRAM
//
// Martin Renters, March/2020
// KIZARM :
// Upraveno tak, aby to šlo přeložit i v RSX11M+ (zkrácení názvů).
// Pak jsou potřeba i soubory font.h, banner.odl, banner.cmd
//////////////////////////////////////////////////////////////////////////*/

#define PAGESIZE  66
#define MAXLINE   80
/* FILL - Centrování na stránky působí na konzoli spíš rušivě */
/*#define FILL*/

/*//////////////////////////////////////////////////////////////////////////
// When compiled on host machine, include stdio and remap console funcs
//////////////////////////////////////////////////////////////////////////*/
#ifdef __pdp11__
#include "console.h"
void cputs (char *s) {
  for (; *s; s++) cputc (*s);
}
void cgets (char *buffer, int size) {
  char c, *p = buffer;
  while (1) {
    c = cgetc();
    if ((c == '\b') || (c == 0x7F)) {
      if (p > buffer) {
        cputs ("\b \b");
        p--;
      } else {
        cputc (7); /* Ring Bell*/
      }
    } else if (c >= ' ') {
      if (p < buffer + size - 2) {
        * (p++) = c;
        cputc (c);
      }
    } else if (c == '\r') {
      cputc (c);
      cputc ('\n');
      *p = '\0';
      return;
    }
  }
}
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void cputs (char *s) {
  while (*s) putchar (* (s++));
}
#define cputc putchar

unsigned char cgetc() {
  unsigned char c = getchar();
  if (c == '\n') c = '\r';
  return c;
}

void cgets (char *buf, int n) {
  char * i;
  i = fgets (buf, n, stdin);
  (void) i;
  buf[strlen (buf)-1] = 0;
}
#endif
/*//////////////////////////////////////////////////////////////////////////
// Include font data generated by mkfont
//////////////////////////////////////////////////////////////////////////*/
#include "font.h"

/*//////////////////////////////////////////////////////////////////////////
// Output buffer
//////////////////////////////////////////////////////////////////////////*/
char outbuf[FONT_MAXWIDTH+3]; /* Account for CR, LF, NUL */
char *obfnex;

/*//////////////////////////////////////////////////////////////////////////
// A bit pointer, includes pointer to byte and a bit mask for the bit
//////////////////////////////////////////////////////////////////////////*/
struct bitpointer {
  unsigned char *byte;
  int bit;
};

/*//////////////////////////////////////////////////////////////////////////
// gbits: get next n bits
//////////////////////////////////////////////////////////////////////////*/
unsigned int gbits (struct bitpointer *bp, int n) {
  unsigned int v = 0;
  unsigned char byte = *bp->byte;
  for (byte = *bp->byte; n>0; n--) {
    v <<= 1;
    if (byte & bp->bit) v |= 1;
    bp->bit >>= 1;
    if (!bp->bit) {
      byte = * (++bp->byte);
      bp->bit = 0x80;
    }
  }
  return v;
}

/*//////////////////////////////////////////////////////////////////////////
// gopcod: get the next opcode. See font.h for details of encoding
//////////////////////////////////////////////////////////////////////////*/
unsigned char gopcod (struct bitpointer *bp) {
  unsigned int v;

  v = gbits (bp, 2);
  if (v == 1) v = gbits (bp, 3) + 1;
  else if (v == 2) v = gbits (bp, 4) + 9;
  else if (v == 3) v = gbits (bp, 6) + 25;

  return flut[v];
}

/*//////////////////////////////////////////////////////////////////////////
// findch: find beginning of character opcodes
//////////////////////////////////////////////////////////////////////////*/
void findch (char c, struct bitpointer *bp) {
  bp->byte = &mfont[foffs[c-32]];
  bp->bit = 0x80;
}
#ifdef FILL
/*//////////////////////////////////////////////////////////////////////////
// lines_for_char: returns count of the number of lines needed for char
//////////////////////////////////////////////////////////////////////////*/
int lines_for_char (char c) {
  struct bitpointer bp;
  int lines = 0;
  unsigned char opcode;

  findch (c, &bp);
  do {
    opcode = gopcod (&bp);
    if (opcode == OPCODE_EOL || opcode == OPCODE_REPEAT)
      lines += 1 + FONT_DOUBLED;
  } while (opcode != OPCODE_EOC);
  lines++;
  return lines;
}
#endif
/*//////////////////////////////////////////////////////////////////////////
// banchr: prints large character
//////////////////////////////////////////////////////////////////////////*/
void banchr (char c) {
  char colors[] = " #";
  struct bitpointer bp;
  int color;
  unsigned char opcode;

  color = 0;
  findch (c, &bp);
  obfnex = outbuf;
  do {
    opcode = gopcod (&bp);
    if (opcode == OPCODE_EOL) {
      * (obfnex++) = '\r';
      * (obfnex++) = '\n';
      * (obfnex++) = 0;
      cputs (outbuf);
      if (FONT_DOUBLED) cputs (outbuf);
      obfnex = outbuf;
      color = 0;
    } else if (opcode == OPCODE_REPEAT) {
      cputs (outbuf);
      if (FONT_DOUBLED) cputs (outbuf);
    } else if (opcode != OPCODE_EOC) {
      while (opcode) {
        * (obfnex++) = colors[color];
        if (FONT_DOUBLED) * (obfnex++) = colors[color];
        opcode--;
      }
      color = (color + 1) & 1; /* No native XOR on PDP11/05 */
    }
  } while (opcode != OPCODE_EOC);
  cputs ("\r\n");
}

/*//////////////////////////////////////////////////////////////////////////
// MAIN ROUTINE
//////////////////////////////////////////////////////////////////////////*/
char buffer[MAXLINE];
#ifdef __pdp11__
void mainf() {
#else
int main() {
#endif
  char *p;
#ifdef FILL
  int i;
  int lines;
#endif
  cputs ("Banner Text? ");
  cgets (buffer, sizeof (buffer));
#ifdef FILL
  /* Count the number of lines the text will take up */
  for (p = buffer, lines = 0; *p; p++) {
    lines += lines_for_char (*p);
  }

  /* Compute remainder for centering. Avoid use of divide since
   some processors (like 11/05) don't have divide instruction */
  for (; lines >= PAGESIZE; lines -= PAGESIZE) ;
  lines = PAGESIZE-lines;

  c_puts ("Align Paper, Press Return\r\n");
  while (cgetc() != '\r') ;

  /* Output leading empty lines */
  for (i=0; i<lines>>1; i++) cputc ('\n');
  lines -= (lines>>1);

  for (p = buffer; *p; p++) {
    banchr (*p);
  }

  /* Output trailing empty lines */
  for (i=0; i<lines; i++) cputc ('\n');
#else
  for (p = buffer; *p; p++) {
    banchr (*p);
  }
#endif
#ifndef __pdp11__
  return 0;
#endif
}
